#include <stdio.h>

#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

void paste(IplImage *src, IplImage *dst, int offx=0, int offy=0)
{
    IplImage *forCopy;
    IplImage *tmp=0;
    if (src->depth!=IPL_DEPTH_8U)
    {
        //need conversion
        tmp=cvCreateImage(cvSize(src->width, src->height), IPL_DEPTH_8U, 1);

        cvConvert(src, tmp);
        forCopy=tmp;
    }
    else
        forCopy=src;

    CvRect selection = cvRect(offx, offy, forCopy->width, forCopy->height);
    cvSetImageROI(dst, selection);
    cvCopy(forCopy, dst);
    cvResetImageROI(dst);

    if (tmp)
        cvReleaseImage(&tmp);
}

void laplacian(IplImage *lu, IplImage *ld, IplImage *out)
{
    IplImage *tmp=cvCreateImage(cvSize(lu->width, lu->height), lu->depth, lu->nChannels);

    cvPyrUp(ld, tmp);

    for(int r=0; r<lu->height; r++)
		for(int c=0;c<lu->width; c++)
		{
			CvScalar v1,v2;			
			v1=cvGet2D(lu, r, c);
            v2=cvGet2D(tmp, r, c);
            double tmpV=(v1.val[0]-v2.val[0]);
			cvSet2D(out, r, c, cvScalar(tmpV));
		}

    cvReleaseImage(&tmp);
}

void add(IplImage *s1, IplImage *s2, IplImage *dest)
{
    for(int r=0; r<s1->height; r++)
		for(int c=0;c<s1->width; c++)
		{
			CvScalar v1,v2;			
			v1=cvGet2D(s1, r, c);
            v2=cvGet2D(s2, r, c);
  
            double tmp=(v2.val[0]+v1.val[0]);
			cvSet2D(dest, r, c, cvScalar(tmp));
		}
}

void drawPyramid(IplImage **p, int L, IplImage *out)
{
    cvSet(out, cvScalar(0,0,0));

    paste(p[0], out, 0, 0);

    int offset=0;
    for(int k=1;k<L;k++)
    {
        paste(p[k], out, p[0]->width, offset);

        offset=offset+p[k]->height;
    }
}

IplImage **allocP(int W, int H, int K, unsigned int DEPTH)
{
    IplImage **ret=new IplImage*[K];
    
    ret[0]=cvCreateImage(cvSize(W, H), DEPTH, 1);

    for(int k=1;k<K;k++)
    {
        ret[k]=cvCreateImage(cvSize(ret[k-1]->width/2, ret[k-1]->height/2), DEPTH, 1);
    }
    return ret;
}

void destroyP(IplImage **p, int K)
{
    for(int k=0; k<K; k++)
    {
        cvReleaseImage(&p[k]);
    }

    delete [] p;
}

void encode(IplImage *img, IplImage **gp, IplImage **lp, int K)
{
    cvCopy(img, gp[0]);
    
    int offset=0;
    for(int k=1;k<K;k++)
    {
       cvPyrDown(gp[k-1],
                 gp[k]);       
       
       laplacian(gp[k-1], gp[k], lp[k-1]);

     }

    //copy last
    cvConvert(gp[K-1], lp[K-1]);
}

void decode(IplImage **lp, IplImage **gp, bool *valid, int K)
{
    cvConvert(lp[K-1], gp[K-1]);

    for(int k=(K-2);k>=0;k--)
    {
        cvPyrUp(gp[k+1], gp[k]);

        if (valid[k])
            add(gp[k], lp[k], gp[k]);
    }
}

int main()
{
	printf("Hello opencv\n");
	
    const int levels = 2;
    IplImage *first;
    IplImage *second;

    first = cvLoadImage("cat1.bmp", CV_LOAD_IMAGE_GRAYSCALE);
    if (first == 0) {
        fprintf(stderr, "Error: cannot load first image\n");
        return 0;
    }

    second = cvLoadImage("cat2.bmp", CV_LOAD_IMAGE_GRAYSCALE); 
    if (second == 0) {
        fprintf(stderr, "Error: cannot load second image\n");
        return 0;
    }

	cvNamedWindow("First");
	cvShowImage("First", first);
    cvWaitKey(0);

    cvNamedWindow("Second");
    cvShowImage("Second", second);
    cvWaitKey(0);

    const int step = 16;
    const int elements = 15; // skip fist and last.
    const int num_features = elements * elements;
    CvPoint2D32f* prev_features = new CvPoint2D32f[num_features];
    int cnt = 0;
    for (int y = 0; y < elements; y++) {
        prev_features[cnt].x = step;
        prev_features[cnt].y = (float) ((y+1) * step);
        cnt++;
        for (int x = 1; x < elements; x++, cnt++) {
            prev_features[cnt].x = prev_features[cnt-1].x + step;
            prev_features[cnt].y = prev_features[cnt-1].y;
        }
    }

    CvPoint2D32f* curr_features = new CvPoint2D32f[num_features];
    int count = 0;
    CvSize win_size;
    win_size.height = 16;
    win_size.width = 16;
    char* status = new char[num_features];
    CvTermCriteria criteria = cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS,20,0.03);
    const int flags = 0;

    cvCalcOpticalFlowPyrLK (first, second, 0, 0,
                            prev_features, curr_features,
                            num_features, win_size, levels, status,
                            0, criteria, flags);

    CvPoint2D32f average;
    average.x = 0;
    average.y = 0;
    cnt = 0;

    CvScalar t;
    t.val[0] = 255.0;
    t.val[1] = 255.0;
    t.val[2] = 255.0;
    t.val[3] = 0.;

    for (int i = 0; i < num_features; i++) {
        if (status[i] == 1) {
            CvPoint node;
            node.x = (int)prev_features[i].x;
            node.y = (int)prev_features[i].y;
            cvCircle(first, node, 3, t);
            node.x = (int)curr_features[i].x;
            node.y = (int)curr_features[i].y;
            cvCircle(second, node, 3, t);
            average.x += (curr_features[i].x - prev_features[i].x);
            average.y += (curr_features[i].y - prev_features[i].y);
            cnt ++;
        }
    }

    average.x /= cnt;
    average.y /= cnt;

    printf("average estimated displacement %f %f\n", average.x, average.y); 

	cvShowImage("First", first);
    cvWaitKey(0);
    cvShowImage("Second", second);
    cvWaitKey(0);

    cvReleaseImage(&first);
    cvReleaseImage(&second);
    delete[] prev_features;
    delete[] curr_features;
    delete[] status;
	cvDestroyWindow("First");
    cvDestroyWindow("Second");

	return 0;
}
